import traceback
import time
import sys
from typing import Dict, Union


class TimerManager(object):
    register = {}

    def __init__(self, container: Union[str, Dict[str, float]] = None):
        self.timers: Dict[str, Timer] = {}
        if isinstance(container, str):
            self.costs = TimerManager.register[container]
        elif container is not None:
            self.costs: Dict[str, float] = container
        else:
            self.costs: Dict[str, float] = {}

    def __getitem__(self, item: str):
        if item not in self.timers:
            self.timers[item] = Timer(item, self.costs)
        if item not in self.costs:
            self.costs[item] = 0.
        return self.timers[item]

    @staticmethod
    def stamp(t: float) -> str:
        r = f'{round(t * 1000 % 1000)}ms'
        t = int(t)
        for n, s in zip(
                [60, 60, 24, 1024],
                ['s', 'm', 'h', 'd'],
        ):
            x = t % n
            t = t // n
            r = f'{x}{s} ' + r
            if t == 0: break
        return r

    @staticmethod
    def regist_container(key: str, container: Dict[str, float]):
        TimerManager.register[key] = container


class Timer(object):
    def __init__(self, key: str, container: Union[str, Dict[str, float]]):
        self.name = key
        self.costs = container

    def __enter__(self):
        if isinstance(self.costs, str):
            self.costs = TimerManager.register[self.costs]
        if self.name not in self.costs:
            self.costs[self.name] = 0.
        self.start = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type or exc_val or exc_tb:
            sys.stderr.write('#%s -------------------------------------- Error Message ---------------------------------------------- \n')
            sys.stderr.write(f'#%s Error: exc_type - {exc_type} \n')
            sys.stderr.write(f'#%s Error: exc_val - {exc_val} \n')
            sys.stderr.write(f'#%s Error: exc_tb - {exc_tb} \n')
            sys.stderr.write('#%s --------------------------------------------------------------------------------------------------- \n')
            sys.stderr.flush()
            traceback.print_exc()
        else:
            self.costs[self.name] += time.time() - self.start
        return False

    def __call__(self, func: callable):
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return wrapper



from .logger import Logger
from .timer import TimerManager, Timer
from .multiprocess import Queue, Closed, Process, QueueMessageException
from .json_encoder import JassorJsonEncoder


__all__ = [
    'Logger',
    'TimerManager', 'Timer',
    'Queue', 'Closed', 'Process', 'QueueMessageException',
    'JassorJsonEncoder',
]



from typing import Union, TextIO
import traceback
import sys
from time import time
from io import TextIOWrapper
from pathlib import Path
from threading import Condition


class IOWrapper:
    def __init__(self, write_func: callable, flush_func: callable, close_func: callable):
        self.write = write_func or self.nothing
        self.flush = flush_func or self.nothing
        self.close = close_func or self.nothing

    @staticmethod
    def nothing(*args, **kwargs):
        pass


class Logger(object):
    STEP = 0
    DEBUG = 1
    INFO = 2
    WARNING = 3

    def __init__(self, start: int = 0, indentation: int = 1, file: Union[TextIO, IOWrapper, TextIOWrapper, str, Path] = sys.stdout, con: Condition = None, level: int = 2):
        self.output_file = file
        self.con = con or Condition()
        self.start = start or time()
        self.indentation = indentation
        self.level = level
        self.block_name = ''
        with self.con:
            if isinstance(self.output_file, (str, Path)):
                self.output_file = open(self.output_file, 'a')

    def close(self):
        with self.con:
            self.output_file.close()

    def track(self, message: str, prefix: str = ''):
        with self.con:
            self.output_file.write('# %s%s %s -> at time %.2f\n' % (prefix, '\t' * self.indentation, message, time() - self.start))
            self.output_file.flush()

    def step(self, message: str):
        if self.level <= Logger.STEP: self.track(message, prefix='STEP')

    def debug(self, message: str):
        if self.level <= Logger.DEBUG: self.track(message, prefix='DEBUG')

    def info(self, message: str):
        if self.level <= Logger.INFO: self.track(message, prefix='INFO')

    def warn(self, message: str):
        if self.level <= Logger.WARNING: self.track(message, prefix='WARN')

    def tab(self):
        with self.con:
            return Logger(start=self.start, indentation=self.indentation+1, file=self.output_file, con=self.con)

    def __getitem__(self, item: str):
        self.block_name = item
        return self

    def __enter__(self):
        self.enter = time()
        with self.con:
            self.track(f'enter {self.block_name} at %.2f seconds' % (time() - self.start), prefix='WITH')
        return self.tab()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type or exc_val or exc_tb:
            sys.stderr.write('#%s -------------------------------------- Error Message ---------------------------------------------- \n')
            sys.stderr.write(f'#%s Error: exc_type - {exc_type} \n')
            sys.stderr.write(f'#%s Error: exc_val - {exc_val} \n')
            sys.stderr.write(f'#%s Error: exc_tb - {exc_tb} \n')
            sys.stderr.write('#%s --------------------------------------------------------------------------------------------------- \n')
            sys.stderr.flush()
            traceback.print_exc()
        else:
            self.track(f'exit {self.block_name} at %.2f seconds -- costing %.2f seconds' % (time() - self.start, time() - self.enter), prefix='EXIT')
        return False
